package ru.dbotthepony.mc.otm.core.util

import it.unimi.dsi.fastutil.ints.IntArrayList
import net.minecraft.util.Mth
import net.minecraft.util.RandomSource
import net.minecraft.world.Container
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.storage.loot.LootParams
import net.minecraft.world.level.storage.loot.LootTable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.container.IEnhancedContainer
import ru.dbotthepony.mc.otm.container.get
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.shuffle
import java.util.stream.Collectors
import java.util.stream.IntStream
import kotlin.math.min

private val LOGGER = LogManager.getLogger()

private class Bucket {
	val items = ArrayList<ItemStack>()
	var last = 0

	fun add(item: ItemStack) {
		if (item.isEmpty) return

		for (i in last until items.size) {
			val existing = items[i]
			var available = existing.maxStackSize - existing.count

			if (available > 0 && ItemStack.isSameItemSameComponents(existing, item)) {
				val diff = min(available, item.count)
				existing.grow(diff)
				item.shrink(diff)
				available -= diff

				if (available == 0 && i == last)
					last++

				if (item.isEmpty)
					return
			}
		}

		if (item.isNotEmpty) {
			items.add(item)
		}
	}
}

private fun recombine(lists: Collection<Collection<ItemStack>>): MutableList<ItemStack> {
	val result = HashMap<Item, Bucket>()

	for (list in lists)
		for (item in list)
			result.computeIfAbsent(item.item) { Bucket() }.add(item)

	return result.values
		.stream()
		.flatMap { it.items.stream() }
		.collect(Collectors.toCollection(::ArrayList))
}

fun LootTable.getRandomItems(params: LootParams, random: RandomSource, rounds: Int = 1): MutableList<ItemStack> {
	if (rounds == 0) {
		return ArrayList()
	} else if (rounds == 1) {
		return getRandomItems(params, random)
	} else {
		return recombine(IntStream.range(0, rounds).mapToObj { getRandomItems(params, random) }.toList())
	}
}

private fun shuffle(items: MutableList<ItemStack>, emptySlotCount: Int, random: RandomSource) {
	val pool = ArrayList<ItemStack>(items)
	items.clear()

	while (items.size + pool.size < emptySlotCount && pool.isNotEmpty()) {
		val select = pool.removeAt(random.nextInt(pool.size))
		val maxStackSize = select.maxStackSize

		if (maxStackSize == 1 || select.count == 1) {
			items.add(select)
		} else {
			val split = select.split(Mth.nextInt(random, 1, select.count / 2))

			if (split.count > 1 && random.nextBoolean())
				pool.add(split)
			else
				items.add(split)

			if (select.count > 1 && random.nextBoolean())
				pool.add(select)
			else
				items.add(select)
		}
	}

	items.addAll(pool)
	items.shuffle(random)
}

fun LootTable.fill(params: LootParams, random: RandomSource, container: Container, rounds: Int = 1) {
	val emptySlots = IntArrayList()

	if (container is IEnhancedContainer<*>) {
		for (i in container.emptySlotIndexIterator())
			emptySlots.add(i)
	} else {
		for (i in 0 until container.containerSize)
			if (container[i].isEmpty)
				emptySlots.add(i)
	}

	emptySlots.shuffle(random)

	if (emptySlots.isEmpty) {
		LOGGER.warn("Tried to fill container with no empty slots")
		return
	}

	val items = getRandomItems(params, random, rounds)
	shuffle(items, emptySlots.size, random)

	val slotItr = emptySlots.iterator()
	val itemItr = items.iterator()

	while (slotItr.hasNext() && itemItr.hasNext()) {
		container[slotItr.nextInt()] = itemItr.next()
	}

	if (itemItr.hasNext()) {
		LOGGER.warn("Tried to overfill a container")
	}
}
